#ifndef NACB_VEC3_H
#define NACB_VEC3_H

#include <math.h>
#include <iostream>
#include <sstream>

#include "nspace.h"

NACB_BEGIN

template <class T>
class Vec3{
public:
#ifdef __PY_PLUS_PLUS
  T x,y,z;
#else
  union{
    struct{
      T x,y,z;
    };
    T data[3];
  };
#endif
  Vec3(const Vec3 & other){
    operator=(other);
  }
  Vec3(T x=0,T y=0,T z=0){
    this->x=x;
    this->y=y;
    this->z=z;
  }
  double dot(const Vec3 & o) const{
    return x*o.x+y*o.y+z*o.z;
  }
  double len() const{
    return sqrt(dot(*this));
  }
  double normalize(){
    double l=this->len();
    if(fabs(l-1)<1e-14)return l;

    if(l>1e-14){
      x/=l;
      y/=l;
      z/=l;
    }
    else{
      x=0;
      y=0;
      z=0;
    }
    return l;
  }
  void operator/=(double d){
    x/=d;
    y/=d;
    z/=d;
  }
  void operator*=(double d){
    x*=d;
    y*=d;
    z*=d;
  }
  Vec3 operator*(const Vec3 & other)const{
    return Vec3(x*other.x,y*other.y,z*other.z);
  }
  Vec3 operator*(double d) const{
    return Vec3(x*d,y*d,z*d);
  }
  void operator+=(const Vec3 & o){
    x+=o.x;
    y+=o.y;
    z+=o.z;
  }
  Vec3 operator+(const Vec3 & o) const{
    return Vec3(x+o.x,y+o.y,z+o.z);
  }
  void operator-=(const Vec3 & o){
    x-=o.x;
    y-=o.y;
    z-=o.z;
  }
  Vec3 operator-(const Vec3 & o) const{
    return Vec3(x-o.x,y-o.y,z-o.z);
  }
  Vec3 operator^(const Vec3 & o)const{
    return Vec3(y*o.z-z*o.y,-x*o.z+z*o.x,x*o.y-y*o.x);
  }
  Vec3 cross(const Vec3 & o)const{return operator^(o);}

  T * getData(){
#ifdef __PY_PLUS_PLUS
    return &x;
#else
    return data;
#endif
  }
  void operator=(const Vec3 & o){
    x=o.x;
    y=o.y;
    z=o.z;
  }
  template <class T2> 
  void operator=(const Vec3<T2> & o){
    x=(T)o.x;
    y=(T)o.y;
    z=(T)o.z;
  }
  Vec3 min(const Vec3 & o) const{
    return Vec3(std::min(x,o.x),std::min(y,o.y),std::min(z,o.z));
  }
  Vec3 max(const Vec3 & o) const{
    return Vec3(std::max(x,o.x),std::max(y,o.y),std::max(z,o.z));
  }
  //operator Vec3<double> () const;

  operator std::string () const{
    std::ostringstream os;
    print(os);
    return os.str();
  }
  void print(std::ostream & ) const;

  T & operator[](int i){
    return getData()[i];
  }
  bool operator==(const Vec3 & o) const{
    return (o.x==x) && (o.y==y) && (o.z==z);
  }
#ifdef __PY_PLUS_PLUS
  void python_set(int i,T t){
    getData()[i]=t;
  }
#endif
};

template <class T>
std::ostream & operator<<(std::ostream & o,const Vec3<T> & v){
  v.print(o);
  return o;
};

template<class T>
Vec3<T> operator*(double d,Vec3<T> & v);

typedef Vec3<double> Vec3d;
typedef Vec3<float> Vec3f;


#define VEC3_INSTANTIATIONS template class Vec3<double>; \
  template class Vec3<float>;   

#ifdef VEC3_INSTANTIATE
VEC3_INSTANTIATIONS
#endif 

NACB_END

#endif
